/* IRCfs - IRC FileServ for *nix.
 * Copyright (C) 2002  Nick 'Zaf' Clifford <zaf@nrc.co.nz>
 * For licensing details, refer to the LICENSE file in the source 
 * code directory.
 * This file will compile by itself (with trace.h) outside of the source tree.
 * So can be used in other projects easily, and also can be used in a
 * debug/testing program. (it also compiles with -pedantic)
 */

#include "trace.h"
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <unistd.h>
#include <sys/select.h>
#include <sys/types.h>
#include <ctype.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

static int trace_write(int fd, const char *buffer, size_t size);
static int trace_write_delay_ohno=0;

static char trace_buffer[1024];

int trace_fd = 2; /* stderr */
unsigned long trace_sys_set = 0xFFFFFFFF;
unsigned long trace_lvl_set = 0xFFFFFFFF;

void __trace(unsigned long sys, unsigned long lvl, 
		int line, const char *file, const char *func,
		const char *format, ...)
{
	va_list vl;
	if (!(sys & trace_sys_set) || !(lvl & trace_lvl_set)) 
		return;

	va_start(vl,format);
	__tracev(sys,lvl,line,file,func,format,vl);
	va_end(vl);
	return;
}

void __tracev(unsigned long sys, unsigned long lvl, 
		int line, const char *file, const char *func,
		const char *format, va_list vl)
{
	char *cp;
	size_t len,t;
	int errno_save = errno;
	if (!(sys & trace_sys_set) || !(lvl & trace_lvl_set)) 
		return;
	
	len = sizeof(trace_buffer)-1;

	snprintf(trace_buffer,len,"[%lx:%lx@%s:%d:%s] ",
			sys,lvl,file,line,func);
	trace_buffer[len]=0;
	t=strlen(trace_buffer);
	cp = trace_buffer + t;
	len-=t;
	
	vsnprintf(cp,len,format,vl);

	len = strlen(trace_buffer);
	if (errno_save) {
		snprintf(trace_buffer+len,sizeof(trace_buffer)-len-1,
				" (errno=%d:%s)",
				errno_save,strerror(errno_save));
		len=strlen(trace_buffer);
	}
	if (len > 0 && trace_buffer[len-1] != '\n' &&
			trace_buffer[len-1] != '\r') {
		trace_buffer[len++] = '\n';
		trace_buffer[len] = 0;
	}

	trace_write(trace_fd,trace_buffer,len);

	if (trace_write_delay_ohno > 0) {
		snprintf(trace_buffer,sizeof(trace_buffer)-1,
				"Trace Write Delay Oh No is %d\n",
				trace_write_delay_ohno);
		trace_write_delay_ohno = 0;
		trace_write(trace_fd,trace_buffer,strlen(trace_buffer));
	}
	errno = errno_save;
	return;

}

void trace_off()
{
	trace_lvl_set = 0x0;
}
void trace_on(unsigned long lvl)
{
	trace_lvl_set = lvl;
}

void trace_sys_on(unsigned long sys)
{
	trace_sys_set = sys;
}
void trace_sys_off()
{
	trace_sys_set = 0;
}

void trace_output_fd(int fd)
{
	trace_fd = fd;
}

#define TRACE_DEFAULT_LVL	T_ALL


static int trace_write_delay(int fd, const char *buffer, size_t size);

static int trace_write(int fd, const char *buffer, size_t size)
{
	int r;
	r = write(fd,buffer,size);
	if (r == -1 && errno == EAGAIN) {
		return trace_write_delay(fd,buffer,size);
	} else if (r == -1) {
		return -1;
	} else if (r == 0) {
		return 0;
	} else if (r != size) {
		return r+trace_write_delay(fd,buffer+r,size-r);
	}
	return r;
}

static int trace_write_delay(int fd, const char *buffer, size_t size)
{
	int r;
	size_t orig_size = size;
	fd_set fdw;
	do {
		trace_write_delay_ohno++;
		r = write(fd,buffer,size);
		if (r == -1 && errno == EAGAIN) {
			FD_ZERO(&fdw);
			FD_SET(fd,&fdw);
			r = select(fd+1,NULL,&fdw,NULL,NULL);
			if (r == -1 && errno == EINTR)
				continue;
			if (r == -1)
				return -1;
			continue;
		}
		if (r == 0)
			return 0;
		buffer+=r;
		size-=r;
	} while(size > 0);
	return orig_size;
}




